import numpy as np
import numpy.random as nrand


def vol(returns) -> float:
    return np.std(returns)


def beta(returns, market) -> float:
    m = np.matrix([returns, market])
    return np.cov(m)[0][1] / np.std(market)


def var(returns, alpha):
    sorted_returns = np.sort(returns)
    index = int(alpha * len(sorted_returns))
    return abs(sorted_returns[index])

def cvar(returns, alpha):
    sorted_returns = np.sort(returns)
    index = int(alpha * len(sorted_returns))
    sum_var = sorted_returns[0]
    for i in range(1, index):
        sum_var += sorted_returns[i]

    return abs(sum_var / index)


def lower_partial_moment(returns, threshold, order):
    threshold_array = numpy.empty(len(returns))
    threshold_array.fill(threshold)

    diff = threshold_array - returns
    diff = diff.clip(min=0)

    return np.sum(diff ** order) / len(returns)


def higher_partial_moment(returns, threshold, order):
    threshold_array = np.empty(len(returns))
    threshold_array.fill(threshold)
    diff = returns - threshold_array
    diff = diff.clip(min=0)
    return np.sum(diff ** order) / len(returns)


def drawdown(returns, tau):
    values = prices(returns, 100)
    pos = len(values) - 1
    pre = pos - tau
    drawdown = float('+inf')

    while pre >= 0:
        dd_i = (values[pos] / values[pre]) - 1
        if dd_i < drawdown:
            drawdown = dd_i

        pos, pre = pos - 1, pre - 1

    return abs(drawdown)

def max_drawdown(returns):
    max_drawdown = float('-inf')
    for i in range(0, len(returns)):
        drawdown_i = dd(returns, i)
        if drawdown_i > max_drawdown:
            max_drawdown = drawdown_i

    return abs(max_drawdown)


def average_drawdown(returns, periods):
    for i in range(0, len(returns)):
        drawdown_i = dd(returns, i)
        drawdowns.append(drawdown_i)
    drawdowns = sorted(drawdowns)
    total_dd = abs(drawdowns[0])
    
    for i in range(1, periods):
        total_dd += abs(drawdowns[i])
    
    return total_dd / periods


def average_drawdown_squared(returns, periods):
    drawdowns = []
    for i in range(0, len(returns)):
        drawdown_i = math.pow(dd(returns, i), 2.0)
        drawdowns.append(drawdown_i)
    drawdowns = sorted(drawdowns)
    total_dd = abs(drawdowns[0])

    for i in range(1, periods):
        total_dd += abs(drawdowns[i])
    return total_dd / periods


def treynor_ratio(er, returns, market, rf):
    return (er - ef) / beta(returns, market)


def sharpe_ratio(er, returns, rf):
    return (er - ef) / vol(returns)


def information_ratio(returns, benchmark):
    diff = returns - benchmark
    return np.mean(diff) / vol(diff)


def modigliani_ratio(er, returns, benchmark, rf):
    np_rf = np.empty(len(returns))
    np_rf.fill(rf)
    rdiff = returns - np_rf
    bdiff = benchmark - np_rf
    return (er - rf) * (vol(rdiff) / vol(bdiff)) + rf


def excess_var(er, returns, rf, alpha):
    return (er - rf) / var(returns, alpha)


def conditional_sharpe_ratio(er, returns, rf, alpha):
    return (er - rf) / cvar(returns, alpha)


def omega_ratio(er, returns, rf, target=0):
    return (er - rf) / lower_partial_moment(returns, target, 1)


def sortino_ratio(er, returns, rf, target=0):
    return (er -rf) / math.sqrt(lower_partial_moment(returns, target, 2))


def kappa_three_ratio(er, returns, rf, target=0):
    return (er - rf) / math.pow(lower_partial_moment(returns, target, 3), float(1/3))


def gain_loss_ratio(returns, target=0):
    return higher_partial_moment(returns, target, 1) / lpm(returns, target, 1)


def upside_potential_ratio(returns, target=0):
    return higher_partial_moment(returns, target, 1) / math.sqrt(lpm(returns, target, 2))


def calmar_ratio(er, returns, rf):
    return (er-rf) / max_drwadown(returns)


def sterling_ratio(er, returns, rf, periods):
    return (er - ef) / average_drawdown(returns, periods)


def burke_ratio(er, returns, rf, periods)
    return (er - ef) / math.sqrt(average_drawdown_squared(returns, periods))


from typing import List
import numpy as np

class cds():
    def __init__(self, num_periods: int, 
        discount_factors: float, 
        default_probabilities: float, 
        recovery_rate: float, 
        notional: float, 
        coupon_frequency: float, 
        issue_premium: float):
        self.num_periods = num_periods
        self.discount_factors = discount_factors
        self.default_probabilities = default_probabilities
        self.recovery_rate = recovery_rate
        self.notional = notional
        self.coupon_frequency = coupon_frequency
        self.issue_premium = issue_premium

    def get_price(self):
        def get_dynamic_price() -> float:
            no_default = np.prod([1 - prob for prob in self.default_probabilities]) * np.sum(self.discount_factors) * \
                (self.issue_premium * self.notional)/ self.coupon_frequency

            return np.sum([
                np.power(1-self.default_probabilities[period], period) * 
                self.default_probabilities[period] * 
                (self.notional * (1 - self.recovery_rate) * self.discount_factors[period] - (self.notional * self.issue_premium * np.sum(self.discount_factors[0:period]))/self.coupon_frequency)
                for period in range(self.num_periods)
            ]) - no_default

        def get_fixed_price() -> float:
            no_default = np.power(1-self.default_probabilities[0], self.num_periods) * \
                self.num_periods * self.discount_factors[0] * (self.notional * self.issue_premium)/self.coupon_frequency

            return np.sum([
                np.power(1-self.default_probabilities[0], period) * 
                self.default_probabilities[0] * 
                (self.notional * (1 - self.recovery_rate) * self.discount_factors [0]- (self.notional * self.issue_premium * self.period)/self.coupon_frequency)
                for period in range(self.num_periods)
            ]) - no_default

        
        if len(self.discount_factors) == 1:
            return get_fixed_price()
        else: 
            return get_dynamic_price()


import pandas as pd


def amortization_table(loan_amount: float, rate: float, monthly_payment: float = 0.):
    table = []
    amount_outstanding = loan_amount
    while amount_outstanding > 0:
        if amount_outstanding == loan_amount:
            table.append(( 
                loan_amount, 
                monthly_payment, 
                loan_amount*(((1+rate)**(1/12)) - 1), 
                monthly_payment-loan_amount*(((1+rate)**(1/12)) - 1)
            ))
            amount_outstanding = loan_amount - monthly_payment-loan_amount*(((1+rate)**(1/12)) - 1)
        else:
            amount_outstanding = table[-1][0] - table[-1][3]
            if amount_outstanding > 0:
                payment = monthly_payment if amount_outstanding > monthly_payment else amount_outstanding*(((1+rate)**(1/12)))

                table.append((
                    amount_outstanding,
                    payment,
                    amount_outstanding*(((1+rate)**(1/12)) - 1),
                    payment-amount_outstanding*(((1+rate)**(1/12)) - 1)
                ))

    return pd.DataFrame(table, columns=['outstanding', 'repayment', 'interest', 'capital_repaid'])





from setuptools import setup

setup(
    name='fin-tools',
    packages=['fin_tools'],
    version='0.1.05',
    description='common financial tools for risk and pricing',
    url='https://github.com/laholmes/fin-tools',
    author='laurence holmes',
    author_email='laholmes@hotmail.co.uk'
)
